from typing import List

from ninja import NinjaAPI

from ledger.api import schema
from ledger.hooks import get_extension_logger
from ledger.models.corporationaudit import CorporationWalletJournalEntry

logger = get_extension_logger(__name__)


class LedgerJournalApiEndpoints:
    tags = ["CorporationJournal"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "corporation/{corporation_id}/wallet/",
            response={200: List[schema.CorporationWalletEvent], 403: str},
            tags=self.tags,
        )
        def get_corporation_wallet(
            request, corporation_id: int, type_refs: str = "", page: int = 1
        ):
            perms = (
                request.user.has_perm("ledger.admin_access")
                | request.user.has_perm("ledger.corp_audit_admin_manager")
                | request.user.has_perm("ledger.corp_audit_manager")
            )

            if not perms:
                return 403, "Permission Denied"

            wallet_journal = (
                CorporationWalletJournalEntry.get_visible(request.user)
                .filter(
                    division__corporation__corporation__corporation_id=corporation_id
                )
                .select_related("first_party", "second_party", "division")
                .order_by("-entry_id")
            )

            start_count = (page - 1) * 10000
            end_count = page * 10000

            if type_refs:
                refs = type_refs.split(",")
                wallet_journal = wallet_journal.filter(ref_type__in=refs)

            wallet_journal = wallet_journal[start_count:end_count]

            output = []
            # pylint: disable=duplicate-code
            for w in wallet_journal:
                output.append(
                    {
                        "division": f"{w.division.division} {w.division.name}",
                        "id": w.entry_id,
                        "date": w.date,
                        "first_party": {
                            "id": w.first_party.eve_id,
                            "name": w.first_party.name,
                            "cat": w.first_party.category,
                        },
                        "second_party": {
                            "id": w.second_party.eve_id,
                            "name": w.second_party.name,
                            "cat": w.second_party.category,
                        },
                        "ref_type": w.ref_type,
                        "amount": w.amount,
                        "balance": w.balance,
                        "reason": w.reason,
                    }
                )

            return output



import datetime
from unittest.mock import MagicMock, patch

import pytz

from django.test import TestCase
from django.utils import timezone
from esi.errors import TokenError
from esi.models import Token

from ledger.errors import DatabaseError
from ledger.models.corporationaudit import CorporationWalletDivision
from ledger.task_helpers.corp_helpers import (
    get_corp_token,
    update_corp_wallet_division,
    update_corp_wallet_journal,
)
from ledger.task_helpers.etag_helpers import NotModifiedError
from ledger.tests.testdata.esi_stub import esi_client_stub
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.task_helpers.corp_helpers"


class GetCorpTokenTest(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()

        cls.mock_token1 = MagicMock(spec=Token)
        cls.mock_token1.character_id = 1001
        cls.mock_token1.valid_access_token.return_value = "token"

        cls.mock_token2 = MagicMock(spec=Token)
        cls.mock_token2.character_id = 1004
        cls.mock_token2.valid_access_token.return_value = "token"

        cls.mock_token3 = MagicMock(spec=Token)
        cls.mock_token3.character_id = 1005
        cls.mock_token3.valid_access_token.return_value = "token"

    @patch(MODULE_PATH + ".EveCharacter.objects.filter")
    @patch(MODULE_PATH + ".Token.objects.filter")
    @patch(MODULE_PATH + ".esi")
    def test_get_corp_token_valid(self, mock_esi, mock_token_filter, mock_char_filter):
        # given
        mock_char_filter.return_value.values.return_value = [
            {"character_id": 1001},
            {"character_id": 1004},
            {"character_id": 1005},
        ]

        mock_tokens = [self.mock_token1, self.mock_token2, self.mock_token3]
        mock_token_filter.return_value.require_scopes.return_value = mock_tokens
        mock_esi.client = esi_client_stub

        # when
        result = get_corp_token(
            2001, ["esi-characters.read_corporation_roles.v1"], ["Factory_Manager"]
        )

        # then
        self.assertEqual(result, self.mock_token1)

    @patch(MODULE_PATH + ".EveCharacter.objects.filter")
    @patch(MODULE_PATH + ".Token.objects.filter")
    @patch(MODULE_PATH + ".esi")
    def test_get_corp_token_valid_norole_scope(
        self, mock_esi, mock_token_filter, mock_char_filter
    ):
        # given
        mock_char_filter.return_value.values.return_value = [
            {"character_id": 1001},
            {"character_id": 1004},
            {"character_id": 1005},
        ]

        mock_tokens = [self.mock_token1, self.mock_token2, self.mock_token3]
        mock_token_filter.return_value.require_scopes.return_value = mock_tokens
        mock_esi.client = esi_client_stub

        # when
        result = get_corp_token(2001, ["No Req Scope"], ["No Role"])

        # then
        self.assertFalse(result)

    @patch(MODULE_PATH + ".EveCharacter.objects.filter")
    @patch(MODULE_PATH + ".Token.objects.filter")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".logger")
    def test_get_corp_token_no_token(
        self, mock_logger, mock_esi, mock_token_filter, mock_char_filter
    ):
        # given
        mock_char_filter.return_value.values.return_value = [
            {"character_id": 1001},
            {"character_id": 1004},
            {"character_id": 1005},
        ]

        mock_tokens = [self.mock_token1, self.mock_token2, self.mock_token3]
        mock_token_filter.return_value.require_scopes.return_value = mock_tokens
        mock_esi.client.Character.get_characters_character_id_roles.side_effect = (
            TokenError()
        )

        # when
        result = get_corp_token(
            2001, ["esi-characters.read_corporation_roles.v1"], ["Factory_Manager"]
        )
        # then
        self.assertFalse(result)
        assert mock_logger.error.called


class UpdateCorpWalletDivisionTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()

        cls.mock_token = MagicMock(spec=Token)
        cls.mock_token.character_id = 1001
        cls.mock_token.corporation_id = 2001
        cls.mock_token.valid_access_token.return_value = "token"
        # TODO: Fetch from Fake ESI
        cls.journal = [
            {
                "amount": 1000,
                "balance": 2000,
                "context_id": 0,
                "context_id_type": "division",
                "date": timezone.now(),
                "description": "Test",
                "first_party_id": 1001,
                "id": 1,
                "entry_id": 1,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1002,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "amount": 1000,
                "balance": 2000,
                "context_id": 0,
                "context_id_type": "division",
                "date": timezone.now(),
                "description": "Test",
                "first_party_id": 1004,
                "id": 2,
                "entry_id": 2,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1002,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "amount": 1000,
                "balance": 2000,
                "context_id": 0,
                "context_id_type": "division",
                "date": timezone.now(),
                "description": "Test",
                "first_party_id": 1010,
                "entry_id": 3,
                "id": 3,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1001,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "amount": 1000,
                "balance": 2000,
                "context_id": 0,
                "context_id_type": "division",
                "date": timezone.now(),
                "description": "Test",
                "first_party_id": 1001,
                "entry_id": 4,
                "id": 4,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1101,
                "tax": 0,
                "tax_receiver_id": 0,
            },
        ]

    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.update_or_create")
    @patch(MODULE_PATH + ".update_corp_wallet_journal")
    def test_update_corp_wallet_division(
        self,
        _,
        mock_update_or_create,
        mock_etag_results,
        mock_esi,
        mock_get_token,
        mock_get_corp,
    ):
        # given
        mock_get_corp.return_value = MagicMock()
        mock_get_corp.return_value.corporation.corporation_id = 2001
        mock_get_token.return_value = self.mock_token
        mock_etag_results.return_value = [{"division": 1, "balance": 100}]
        mock_update_or_create.return_value = (MagicMock(), True)
        mock_esi.client = esi_client_stub
        # when
        result = update_corp_wallet_division(2001)
        # then
        self.assertEqual(
            result,
            (
                "Finished wallet divs for: %s",
                mock_get_corp.return_value.corporation.corporation_name,
            ),
        )

    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.update_or_create")
    @patch(MODULE_PATH + ".update_corp_wallet_journal")
    def test_update_corp_wallet_division_no_division_item(
        self,
        _,
        mock_update_or_create,
        mock_etag_results,
        mock_esi,
        mock_get_token,
        mock_get_corp,
    ):
        # given
        mock_get_corp.return_value = MagicMock()
        mock_get_corp.return_value.corporation.corporation_id = 2001
        mock_get_token.return_value = self.mock_token
        mock_etag_results.return_value = [{"division": 1, "balance": 100}]
        mock_update_or_create.return_value = (None, False)
        mock_esi.client = esi_client_stub
        # when
        result = update_corp_wallet_division(2001)
        # then
        self.assertEqual(
            result,
            (
                "Finished wallet divs for: %s",
                mock_get_corp.return_value.corporation.corporation_name,
            ),
        )

    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.update_or_create")
    @patch(MODULE_PATH + ".update_corp_wallet_journal")
    def test_update_corp_wallet_division_no_token(
        self,
        _,
        mock_update_or_create,
        mock_etag_results,
        mock_esi,
        mock_get_token,
        mock_get_corp,
    ):
        # given
        mock_get_corp.return_value = MagicMock()
        mock_get_corp.return_value.corporation.corporation_id = 2001
        mock_get_token.return_value = None
        mock_etag_results.return_value = [{"division": 1, "balance": 100}]
        mock_update_or_create.return_value = (MagicMock(), True)
        mock_esi.client = esi_client_stub
        # when
        result = update_corp_wallet_division(2001)
        # then
        self.assertEqual(result, "No Tokens")

    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.update_or_create")
    @patch(MODULE_PATH + ".update_corp_wallet_journal")
    @patch(MODULE_PATH + ".logger")
    def test_update_corp_wallet_division_not_modified(
        self,
        mock_logger,
        _,
        mock_update_or_create,
        mock_etag_results,
        mock_esi,
        mock_get_token,
        mock_get_corp,
    ):
        # given
        mock_get_corp.return_value = MagicMock()
        mock_get_corp.return_value.corporation.corporation_id = 2001
        mock_get_token.return_value = self.mock_token
        mock_etag_results.side_effect = NotModifiedError()
        mock_update_or_create.return_value = (MagicMock(), True)
        mock_esi.client = esi_client_stub
        # when
        result = update_corp_wallet_division(2001)
        # then
        self.assertEqual(
            result,
            (
                "Finished wallet divs for: %s",
                mock_get_corp.return_value.corporation.corporation_name,
            ),
        )
        mock_logger.debug.assert_called_once_with(
            "No New wallet data for: %s",
            mock_get_corp.return_value.corporation.corporation_name,
        )

    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletJournalEntry.objects.filter")
    @patch(MODULE_PATH + ".EveEntity.objects.all")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EveEntity.objects.create_bulk_from_esi")
    @patch(MODULE_PATH + ".CorporationWalletJournalEntry.objects.bulk_create")
    def test_update_corp_wallet_journal_success(
        self,
        mock_bulk_create,
        mock_create_bulk_from_esi,
        mock_etag_results,
        mock_esi,
        mock_all,
        mock_filter,
        mock_division_get,
        mock_audit_get,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token

        mock_audit_get.return_value = MagicMock()
        mock_audit_get.return_value.corporation.corporation_id = 2001

        mock_division_get.return_value = CorporationWalletDivision(
            name=None, balance=1000.00, division=1
        )

        mock_filter.return_value.order_by.return_value.values_list.return_value = [1, 3]

        mock_all.return_value.values_list.return_value = []

        mock_esi.client = esi_client_stub

        date_string = "2016-10-29T14:00:00Z"
        date_object = datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
        date_object = date_object.replace(tzinfo=pytz.UTC)

        mock_etag_results.return_value = self.journal

        mock_create_bulk_from_esi.return_value = True
        mock_bulk_create.return_value = [MagicMock()]

        # when
        result = update_corp_wallet_journal(2001, 1)

        # then
        self.assertEqual(result, True)

    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.get")
    def test_update_corp_wallet_journal_no_token(
        self, mock_division_get, mock_audit_get, mock_get_token
    ):
        # given
        mock_get_token.return_value = None
        mock_audit_get.return_value = MagicMock()
        mock_division_get.return_value = MagicMock()

        # when
        result = update_corp_wallet_journal(2001, 1)

        # then
        self.assertEqual(result, "No Tokens")

    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletJournalEntry.objects.filter")
    @patch(MODULE_PATH + ".EveEntity.objects.all")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".logger")
    def test_update_corp_wallet_journal_not_modified(
        self,
        mock_logger,
        mock_etag_results,
        mock_esi,
        mock_all,
        mock_filter,
        mock_division_get,
        mock_audit_get,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token

        mock_audit_get.return_value = MagicMock()
        mock_audit_get.return_value.corporation.corporation_id = 2001

        mock_division_get.return_value = CorporationWalletDivision(
            name=None, balance=1000.00, division=1
        )

        mock_filter.return_value.order_by.return_value.values_list.return_value = []

        mock_all.return_value.values_list.return_value = []

        mock_esi.client = esi_client_stub

        date_string = "2016-10-29T14:00:00Z"
        date_object = datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
        date_object = date_object.replace(tzinfo=pytz.UTC)

        mock_etag_results.side_effect = NotModifiedError()

        # when
        result = update_corp_wallet_journal(2001, 1)

        # then
        self.assertEqual(result, True)
        mock_logger.debug.assert_called_with(
            "No New wallet data for: Div: %s Corp: %s",
            mock_audit_get.return_value.corporation.corporation_name,
            1,
        )

    @patch(MODULE_PATH + ".get_corp_token")
    @patch(MODULE_PATH + ".CorporationAudit.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletDivision.objects.get")
    @patch(MODULE_PATH + ".CorporationWalletJournalEntry.objects.filter")
    @patch(MODULE_PATH + ".EveEntity.objects.all")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EveEntity.objects.create_bulk_from_esi")
    def test_update_corp_wallet_journal_raise_db(
        self,
        mock_create_bulk_from_esi,
        mock_etag_results,
        mock_esi,
        mock_all,
        mock_filter,
        mock_division_get,
        mock_audit_get,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token

        mock_audit_get.return_value = MagicMock()
        mock_audit_get.return_value.corporation.corporation_id = 2001

        mock_division_get.return_value = CorporationWalletDivision(
            name=None, balance=1000.00, division=1
        )

        mock_filter.return_value.order_by.return_value.values_list.return_value = []

        mock_all.return_value.values_list.return_value = []

        mock_esi.client = esi_client_stub

        date_string = "2016-10-29T14:00:00Z"
        date_object = datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
        date_object = date_object.replace(tzinfo=pytz.UTC)

        mock_etag_results.return_value = self.journal

        mock_create_bulk_from_esi.return_value = None  # No names created

        # when
        with self.assertRaises(DatabaseError):
            result = update_corp_wallet_journal(2001, 1)

            # then
            self.assertEqual(result, True)



from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory, TestCase

from allianceauth.eveonline.evelinks import eveimageserver
from app_utils.testdata_factories import UserFactory

from ledger.admin import CharacterAuditAdmin, CorporationAuditAdmin
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class MockRequest:
    pass


class TestCorporationAuditAdmin(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.factory = RequestFactory()
        cls.site = AdminSite()
        cls.corporation_audit_admin = CorporationAuditAdmin(CorporationAudit, cls.site)
        cls.corporation_audit = CorporationAudit.objects.get(id=1)

    def test_entity_pic(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "corporation", self.corporation_audit.corporation.corporation_id, 32
            )
        )
        self.assertEqual(
            self.corporation_audit_admin._entity_pic(self.corporation_audit),
            expected_html,
        )

    def test_corporation_corporation_id(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.corporation_audit_admin._corporation__corporation_id(
                self.corporation_audit
            ),
            2001,
        )

    def test_last_update_wallet(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertEqual(
            self.corporation_audit_admin._last_update_wallet(self.corporation_audit),
            None,
        )

    def test_has_add_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        self.assertFalse(self.corporation_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        request = self.factory.get("/")
        request.user = user
        request = MockRequest()
        self.assertFalse(self.corporation_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.corporation_audit_admin.has_change_permission(
                request, obj=self.corporation_audit
            )
        )


class TestCharacterAuditAdmin(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        # Assuming load_allianceauth() and load_char_audit() are functions you've defined elsewhere to set up your test environment
        load_allianceauth()
        load_ledger_all()
        cls.factory = RequestFactory()
        cls.site = AdminSite()
        cls.character_audit_admin = CharacterAuditAdmin(CharacterAudit, cls.site)
        cls.character_audit = CharacterAudit.objects.get(id=1)

    def test_entity_pic(self):
        user = UserFactory(is_superuser=True, is_staff=True)
        self.client.force_login(user)
        expected_html = '<img src="{}" class="img-circle">'.format(
            eveimageserver._eve_entity_image_url(
                "character", self.character_audit.character.character_id, 32
            )
        )
        self.assertEqual(
            self.character_audit_admin._entity_pic(self.character_audit), expected_html
        )

    def test_character_character_name(self):
        self.assertEqual(
            self.character_audit_admin._character__character_name(self.character_audit),
            self.character_audit.character.character_name,
        )

    def test_last_update_wallet(self):
        self.assertEqual(
            self.character_audit_admin._last_update_wallet(self.character_audit),
            self.character_audit.last_update_wallet,
        )

    def test_last_update_mining(self):
        self.assertEqual(
            self.character_audit_admin._last_update_mining(self.character_audit),
            self.character_audit.last_update_mining,
        )

    def test_has_add_permission(self):
        request = self.factory.get("/")
        self.assertFalse(self.character_audit_admin.has_add_permission(request))

    def test_has_change_permission(self):
        request = self.factory.get("/")
        self.assertFalse(self.character_audit_admin.has_change_permission(request))
        self.assertFalse(
            self.character_audit_admin.has_change_permission(
                request, obj=self.character_audit
            )
        )



# Generated by Django 4.2.11 on 2024-07-13 15:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("ledger", "0002_delete_corpsteuer_alter_characteraudit_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="characteraudit",
            options={
                "default_permissions": (),
                "permissions": (
                    ("char_audit_manager", "Has access to all characters for own Corp"),
                    ("char_audit_admin_manager", "Has access to all characters"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="corporationaudit",
            options={
                "default_permissions": (),
                "permissions": (
                    ("corp_audit_manager", "Has access to own Corporation"),
                    ("corp_audit_admin_manager", "Has access to all Corporations"),
                ),
            },
        ),
    ]



from unittest.mock import MagicMock, patch

from django.db.models import Q
from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from ledger.managers.characterjournal_manager import CharWalletManager
from ledger.models.characteraudit import CharacterWalletJournalEntry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.characterjournal_manager"


class CharManagerQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.manager = CharacterWalletJournalEntry.objects.all()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )

    def test_annotate_bounty(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_bounty(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_bounty", qs.query.annotations)

        qs_no_filter = self.manager.annotate_bounty(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_bounty", qs_no_filter.query.annotations)

    def test_annotate_ess(self):
        character_ids = [1, 2, 3]

        qs = self.manager.filter_ess(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_ess", qs.query.annotations)

        qs_no_filter = self.manager.filter_ess(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_ess", qs_no_filter.query.annotations)

    def test_annotate_mission(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_mission(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_mission", qs.query.annotations)

        qs_no_filter = self.manager.annotate_mission(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_mission", qs_no_filter.query.annotations)

    def test_annotate_mining(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_mining(character_ids)
        self.assertIsNotNone(qs)

        qs_no_filter = self.manager.annotate_mining(character_ids)
        self.assertIsNotNone(qs_no_filter)

    def test_annotate_contract_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_contract_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_contract_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_contract_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_contract_cost", qs_no_filter.query.annotations)

    def test_annotate_market_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_market_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_market_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_market_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_market_cost", qs_no_filter.query.annotations)

    def test_annotate_assets_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_assets_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_assets_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_assets_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_assets_cost", qs_no_filter.query.annotations)

    def test_annotate_traveling_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_traveling_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_traveling_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_traveling_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_traveling_cost", qs_no_filter.query.annotations)

    def test_annotate_production_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_production_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_production_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_production_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_production_cost", qs_no_filter.query.annotations)

    def test_annotate_skill_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_skill_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_skill_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_skill_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_skill_cost", qs_no_filter.query.annotations)

    def test_annotate_insurance_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_insurance_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_insurance_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_insurance_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_insurance_cost", qs_no_filter.query.annotations)

    def test_annotate_planetary_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_planetary_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_planetary_cost", qs.query.annotations)

        qs_no_filter = self.manager.annotate_planetary_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_planetary_cost", qs_no_filter.query.annotations)

    def test_annotate_lp_cost(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_lp_cost(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_lp", qs.query.annotations)

        qs_no_filter = self.manager.annotate_lp_cost(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_lp", qs_no_filter.query.annotations)

    def test_annotate_market_trade(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_market_trade(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_market_trade", qs.query.annotations)

        qs_no_filter = self.manager.annotate_market_trade(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_market_trade", qs_no_filter.query.annotations)

    def test_annotate_contract_trade(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_contract_trade(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_contract_trade", qs.query.annotations)

        qs_no_filter = self.manager.annotate_contract_trade(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_contract_trade", qs_no_filter.query.annotations)

    def test_annotate_donation_trade(self):
        character_ids = [1, 2, 3]
        exclude = [4, 5]

        qs = self.manager.annotate_donation_trade(character_ids, exclude)
        self.assertIsNotNone(qs)
        self.assertIn("total_donation_trade", qs.query.annotations)

        qs_no_filter = self.manager.annotate_donation_trade(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_donation_trade", qs_no_filter.query.annotations)

    def test_annotate_insurance_trade(self):
        character_ids = [1, 2, 3]

        qs = self.manager.annotate_insurance_trade(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_insurance_trade", qs.query.annotations)

        qs_no_filter = self.manager.annotate_insurance_trade(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_insurance_trade", qs_no_filter.query.annotations)

    def test_generate_ledger(self):

        character_ids = [1, 2, 3]
        filter_date = Q(date__gte="2023-01-01")
        exclude = [4, 5]

        # Test with filter
        result_with_filter = self.manager.generate_ledger(
            character_ids, filter_date, exclude
        )
        self.assertIsNotNone(result_with_filter)
        self.assertIn("amounts", result_with_filter)
        self.assertIn("amounts_others", result_with_filter)
        self.assertIn("amounts_costs", result_with_filter)

        # Test without filter
        result_without_filter = self.manager.generate_ledger(character_ids)
        self.assertIsNotNone(result_without_filter)
        self.assertIn("amounts", result_without_filter)
        self.assertIn("amounts_others", result_without_filter)
        self.assertIn("amounts_costs", result_without_filter)

    def test_generate_billboard(self):

        character_ids = [1, 2, 3]
        filter_date = Q(date__gte="2023-01-01")
        exclude = [4, 5]

        # Test with filter
        qs = self.manager.generate_billboard(character_ids, filter_date, exclude)
        self.assertIsNotNone(qs)
        self.assertIn("total_bounty", qs.query.annotations)

        # Test without filter
        qs_no_filter = self.manager.generate_billboard(character_ids)
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("total_bounty", qs_no_filter.query.annotations)



from decimal import Decimal

from django.db import models
from django.db.models import DecimalField, Q, Sum, Value
from django.db.models.functions import Coalesce

from ledger.hooks import get_extension_logger
from ledger.view_helpers.core import events_filter

logger = get_extension_logger(__name__)
# PvE - Income
BOUNTY_PRIZES = ["bounty_prizes"]
ESS_TRANSFER = ["ess_escrow_transfer"]
MISSION_REWARD = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
# Cost Ref Types
# pylint: disable=duplicate-code
CONTRACT_COST = [
    "contract_price_payment_corp",
    "contract_reward",
    "contract_price",
    "contract_collateral",
    "contract_reward_deposited",
]
MARKET_COST = ["market_escrow", "transaction_tax", "market_provider_tax", "brokers_fee"]
ASSETS_COST = ["asset_safety_recovery_tax"]
TRAVELING_COST = ["structure_gate_jump", "jump_clone_activation_fee"]
PRODUCTION_COST = [
    "industry_job_tax",
    "manufacturing",
    "researching_time_productivity",
    "researching_material_productivity",
    "copying",
    "reprocessing_tax",
    "reaction",
]
SKILL_COST = ["skill_purchase"]
INSURANCE_COST = ["insurance"]
PLANETARY_COST = [
    "planetary_export_tax",
    "planetary_import_tax",
    "planetary_construction",
]
LP_COST = ["lp_store"]
# Trading
MARKET_TRADE = ["market_transaction"]
CONTRACT_TRADE = ["contract_price_payment_corp", "contract_reward", "contract_price"]
DONATION_TRADE = ["player_donation"]
INSURANCE_TRADE = ["insurance"]


class CharWalletQuerySet(models.QuerySet):
    # PvE - Income
    def annotate_bounty(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_bounty=Coalesce(
                Sum(
                    "amount",
                    Q(ref_type__in=BOUNTY_PRIZES, second_party_id__in=character_ids),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def filter_ess(self, character_ids: list, filter_date=None) -> models.QuerySet:
        # pylint: disable=import-outside-toplevel
        from ledger.models.corporationaudit import CorporationWalletJournalEntry

        qs = CorporationWalletJournalEntry.objects.get_queryset()

        if filter_date:
            qs = qs.filter(filter_date)

        qs = events_filter(qs)

        qs = qs.filter(ref_type__in=ESS_TRANSFER, second_party_id__in=character_ids)

        qs = qs.annotate(
            total_ess=Coalesce(Sum("amount"), Value(0), output_field=DecimalField())
        )
        return qs

    def annotate_mission(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_mission=Coalesce(
                Sum(
                    "amount",
                    Q(
                        ref_type__in=MISSION_REWARD,
                        second_party_id__in=character_ids,
                        amount__gt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_mining(self, character_ids: list, filter_date=None) -> models.QuerySet:
        # pylint: disable=import-outside-toplevel
        from ledger.models.characteraudit import CharacterMiningLedger

        qs = CharacterMiningLedger.objects.get_queryset()

        qs = qs.filter(character__character__character_id__in=character_ids)

        if filter_date:
            qs = qs.filter(filter_date)

        return qs.annotate_pricing().values(
            "character__character__character_id", "total"
        )

    # Costs
    def annotate_contract_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_contract_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=CONTRACT_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_market_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_market_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=MARKET_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_assets_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_assets_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=ASSETS_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_traveling_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_traveling_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=TRAVELING_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_production_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_production_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=PRODUCTION_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_skill_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_skill_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=SKILL_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_insurance_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=INSURANCE_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_planetary_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_planetary_cost=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=PLANETARY_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_lp_cost(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_lp=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=LP_COST,
                        amount__lt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # Trading - Income
    def annotate_market_trade(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_market_trade=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=MARKET_TRADE,
                        amount__gt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_contract_trade(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_contract_trade=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=CONTRACT_TRADE,
                        amount__gt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_donation_trade(
        self, character_ids: list, exclude=None
    ) -> models.QuerySet:
        qs = self

        if exclude:
            qs = qs.exclude(first_party_id__in=exclude)

        return qs.annotate(
            total_donation_trade=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=DONATION_TRADE,
                        amount__gt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_trade(self, character_ids: list) -> models.QuerySet:
        return self.annotate(
            total_insurance_trade=Coalesce(
                Sum(
                    "amount",
                    Q(
                        Q(first_party_id__in=character_ids)
                        | Q(second_party_id__in=character_ids),
                        ref_type__in=INSURANCE_TRADE,
                        amount__gt=0,
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # Supports Member Audit Intigration
    def generate_ledger(
        self, character_ids: list, filter_date=None, exclude=None
    ) -> models.QuerySet:
        """Generate the Ledger for the given characters."""

        # Filter Characters
        qs = self.filter(
            Q(first_party_id__in=character_ids) | Q(second_party_id__in=character_ids)
        )

        if filter_date:
            qs = qs.filter(filter_date)

        # Aggregate the results directly from the original fields
        ledger_data = qs.aggregate(
            total_bounty=Sum("amount", filter=Q(ref_type__in=BOUNTY_PRIZES)),
            total_mission=Sum("amount", filter=Q(ref_type__in=MISSION_REWARD)),
            total_contract_cost=Sum(
                "amount", filter=Q(ref_type__in=CONTRACT_COST, amount__lt=0)
            ),
            total_market_cost=Sum(
                "amount", filter=Q(ref_type__in=MARKET_COST, amount__lt=0)
            ),
            total_assets_cost=Sum(
                "amount", filter=Q(ref_type__in=ASSETS_COST, amount__lt=0)
            ),
            total_traveling_cost=Sum(
                "amount", filter=Q(ref_type__in=TRAVELING_COST, amount__lt=0)
            ),
            total_production_cost=Sum(
                "amount", filter=Q(ref_type__in=PRODUCTION_COST, amount__lt=0)
            ),
            total_skill_cost=Sum(
                "amount", filter=Q(ref_type__in=SKILL_COST, amount__lt=0)
            ),
            total_insurance_cost=Sum(
                "amount", filter=Q(ref_type__in=INSURANCE_COST, amount__lt=0)
            ),
            total_planetary_cost=Sum(
                "amount", filter=Q(ref_type__in=PLANETARY_COST, amount__lt=0)
            ),
            total_lp=Sum("amount", filter=Q(ref_type__in=LP_COST, amount__lt=0)),
            total_market_trade=Sum(
                "amount", filter=Q(ref_type__in=MARKET_TRADE, amount__gt=0)
            ),
            total_contract_trade=Sum(
                "amount", filter=Q(ref_type__in=CONTRACT_TRADE, amount__gt=0)
            ),
            total_insurance_trade=Sum(
                "amount", filter=Q(ref_type__in=INSURANCE_TRADE, amount__gt=0)
            ),
            total_donation_trade=Sum(
                "amount",
                filter=(
                    Q(ref_type__in=DONATION_TRADE, amount__gt=0)
                    & ~Q(first_party_id__in=exclude)
                    if exclude is not None
                    else Q(ref_type__in=DONATION_TRADE, amount__gt=0)
                ),
            ),
        )

        # Special cases
        ess_data = self.filter_ess(character_ids, filter_date).aggregate(
            total_ess=Sum("amount")
        )
        mining_data = self.annotate_mining(character_ids, filter_date).aggregate(
            total_mining=Sum("total")
        )

        amounts = {
            "bounty": Decimal(ledger_data["total_bounty"] or 0),
            "ess": Decimal(ess_data["total_ess"] or 0),
            "mining": Decimal(mining_data["total_mining"] or 0),
        }

        amounts_others = {
            "mission": Decimal(ledger_data["total_mission"] or 0),
            "donation": Decimal(ledger_data["total_donation_trade"] or 0),
            "transaction": Decimal(ledger_data["total_market_trade"] or 0),
            "contract": Decimal(ledger_data["total_contract_trade"] or 0),
            "insurance": Decimal(ledger_data["total_insurance_trade"] or 0),
        }

        amounts_costs = {
            "contract_cost": Decimal(ledger_data["total_contract_cost"] or 0),
            "market_cost": Decimal(ledger_data["total_market_cost"] or 0),
            "assets_cost": Decimal(ledger_data["total_assets_cost"] or 0),
            "traveling_cost": Decimal(ledger_data["total_traveling_cost"] or 0),
            "production_cost": Decimal(ledger_data["total_production_cost"] or 0),
            "skill_cost": Decimal(ledger_data["total_skill_cost"] or 0),
            "insurance_cost": Decimal(ledger_data["total_insurance_cost"] or 0),
            "planetary_cost": Decimal(ledger_data["total_planetary_cost"] or 0),
            "lp_cost": Decimal(ledger_data["total_lp"] or 0),
        }

        return {
            "amounts": amounts,
            "amounts_others": amounts_others,
            "amounts_costs": amounts_costs,
        }

    # Not Implemented
    def generate_billboard(
        self, character_ids: list, filter_date=None, exclude=None
    ) -> models.QuerySet:
        """Generate the Billboard for the given characters."""
        # Combine all filters
        qs = self.filter(
            Q(first_party_id__in=character_ids) | Q(second_party_id__in=character_ids)
        )

        if filter_date:
            qs = qs.filter(filter_date)

        # Annotate all fields except ESS, Mining, and Donation Trade
        qs = qs.annotate(
            total_bounty=Coalesce(
                Sum("amount", filter=Q(ref_type__in=BOUNTY_PRIZES)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_mission=Coalesce(
                Sum("amount", filter=Q(ref_type__in=MISSION_REWARD)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_contract_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=CONTRACT_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_market_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=MARKET_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_assets_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=ASSETS_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_traveling_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=TRAVELING_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_production_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=PRODUCTION_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_skill_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=SKILL_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_insurance_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=INSURANCE_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_planetary_cost=Coalesce(
                Sum("amount", filter=Q(ref_type__in=PLANETARY_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_lp=Coalesce(
                Sum("amount", filter=Q(ref_type__in=LP_COST, amount__lt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_market_trade=Coalesce(
                Sum("amount", filter=Q(ref_type__in=MARKET_TRADE, amount__gt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_contract_trade=Coalesce(
                Sum("amount", filter=Q(ref_type__in=CONTRACT_TRADE, amount__gt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_insurance_trade=Coalesce(
                Sum("amount", filter=Q(ref_type__in=INSURANCE_TRADE, amount__gt=0)),
                Value(0),
                output_field=DecimalField(),
            ),
            total_donation_trade=Coalesce(
                Sum(
                    "amount",
                    filter=(
                        Q(ref_type__in=DONATION_TRADE, amount__gt=0)
                        & ~Q(first_party_id__in=exclude)
                        if exclude is not None
                        else Q(ref_type__in=DONATION_TRADE, amount__gt=0)
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            ),
        )

        return qs


class CharWalletManagerBase(models.Manager):
    pass


CharWalletManager = CharWalletManagerBase.from_queryset(CharWalletQuerySet)



from celery_once import AlreadyQueued

from django.core.exceptions import ObjectDoesNotExist

from allianceauth.eveonline.models import EveCharacter

from ledger import app_settings, models
from ledger.hooks import get_corp_models_and_string, get_extension_logger
from ledger.tasks import create_missing_character

logger = get_extension_logger(__name__)


def convert_ess_payout(ess: int) -> float:
    """Convert ESS payout"""
    return (ess / app_settings.LEDGER_CORP_TAX) * (100 - app_settings.LEDGER_CORP_TAX)


def get_character(request, character_id):
    """Get Character and check permissions"""
    perms = True
    if character_id == 0:
        character_id = request.user.profile.main_character.character_id

    try:
        main_char = EveCharacter.objects.get(character_id=character_id)
    except ObjectDoesNotExist:
        main_char = EveCharacter.objects.select_related(
            "character_ownership",
            "character_ownership__user__profile",
            "character_ownership__user__profile__main_character",
        ).get(character_id=request.user.profile.main_character.character_id)

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        perms = False
    return perms, main_char


def get_corporation(request, corporation_id):
    """Get Corporation and check permissions"""
    perms = True
    if corporation_id == 0:
        corporations = get_main_and_alts_corporations(request)
    else:
        corporations = [corporation_id]

    main_corp = models.CorporationAudit.objects.filter(
        corporation__corporation_id__in=corporations
    )
    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)
    # Check if there is an intersection between main_corp and visible
    common_corps = main_corp.intersection(visible)
    if not common_corps.exists():
        perms = False
    return perms, main_corp


def get_alts_queryset(main_char, corporations=None):
    """Get all alts for a main character, optionally filtered by corporations."""
    try:
        linked_corporations = (
            main_char.character_ownership.user.character_ownerships.all()
        )

        if corporations:
            linked_corporations = linked_corporations.filter(
                character__corporation_id__in=corporations
            )

        linked_corporations = linked_corporations.values_list("character_id", flat=True)

        return EveCharacter.objects.filter(id__in=linked_corporations)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def _get_linked_characters(corporations):
    linked_chars = EveCharacter.objects.filter(corporation_id__in=corporations)
    linked_chars |= EveCharacter.objects.filter(
        character_ownership__user__profile__main_character__corporation_id__in=corporations
    )
    return (
        linked_chars.select_related(
            "character_ownership", "character_ownership__user__profile__main_character"
        )
        .prefetch_related("character_ownership__user__character_ownerships")
        .order_by("character_name")
    )


def _process_character(
    char: EveCharacter, characters, chars_list, corporations, missing_chars
):
    try:
        main = char.character_ownership.user.profile.main_character
        if main and main.character_id not in characters:
            characters[main.character_id] = {"main": main, "alts": []}
        if char.corporation_id in corporations:
            characters[main.character_id]["alts"].append(char)
            chars_list.add(char.character_id)
    except ObjectDoesNotExist:
        if EveCharacter.objects.filter(character_id=char.character_id).exists():
            char = EveCharacter.objects.get(character_id=char.character_id)
            characters[char.character_id] = {"main": char, "alts": []}
            if char.corporation_id in corporations:
                chars_list.add(char.character_id)
                characters[char.character_id]["alts"].append(char)

        missing_chars.add(char.character_id)
    except AttributeError:
        pass


def _process_missing_characters(missing_chars):
    if missing_chars:
        try:
            create_missing_character.apply_async(args=[list(missing_chars)], priority=6)
        except AlreadyQueued:
            pass


def get_main_and_alts_all(corporations: list):
    """Get all members for given corporations"""
    characters = {}
    chars_list = set()
    missing_chars = set()

    linked_chars = _get_linked_characters(corporations)
    corpmember = get_corp_models_and_string()

    for char in linked_chars:
        _process_character(char, characters, chars_list, corporations, missing_chars)

    for member in corpmember.objects.filter(
        corpstats__corp__corporation_id__in=corporations
    ).exclude(character_id__in=chars_list):
        try:
            char = EveCharacter.objects.get(character_id=member.character_id)
            _process_character(
                char, characters, chars_list, corporations, missing_chars
            )
        except ObjectDoesNotExist:
            missing_chars.add(member.character_id)

    _process_missing_characters(missing_chars)

    return characters, list(chars_list)


def get_main_and_alts_corporations(request):
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.select_related(
        "character", "user"
    ).all()

    linked_characters = linked_characters.values_list("character_id", flat=True)
    chars = EveCharacter.objects.filter(id__in=linked_characters)

    corporations = set()

    for char in chars:
        corporations.add(char.corporation_id)

    return list(corporations)



from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase
from eveuniverse.models import EveMarketPrice

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.character.ledger import LedgerApiEndpoints
from ledger.api.managers.character_manager import CharacterProcess
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.tests.test_api._ledgerchardata import (
    CharmonthlyMarch,
    CharmonthlyMarchMulti,
    CharNoMining,
    Charyearly,
    noData,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.api.managers.character_manager"


class ManageApiLedgerCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.char_audit_manager",
                "ledger.char_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    @patch(MODULE_PATH + ".get_alts_queryset")
    def test_get_alts_with_exception(self, mock_get_alts_queryset):
        process = CharacterProcess(chars=[], year=2024, month=3)
        main = [1]
        mock_get_alts_queryset.side_effect = Exception("Test exception")
        result = process.get_alts(main)
        self.assertIsNone(result)
        mock_get_alts_queryset.assert_called_once_with(1)

    def test_get_character_ledger_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_multi(self):
        chars = EveCharacter.objects.filter(character_id__in=[1002, 1003])
        for char in chars:
            add_character_to_user(self.user, char)
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarchMulti
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_cache_no_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_no_cache_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_no_cache_no_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/ledger/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_year(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/ledger/year/2024/month/0/"
        expected_data = Charyearly
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/1001/ledger/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_character_ledger_api_no_data(self):
        # given
        self.client.force_login(self.user3)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_no_mining(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/year/2024/month/3/"
        CharacterMiningLedger.objects.all().delete()
        EveMarketPrice.objects.all().delete()
        # when
        response = self.client.get(url)
        # then
        expected_data = CharNoMining
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_admin(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/ledger/admin/"
        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "character": {
                    "1002": {
                        "character_id": 1002,
                        "character_name": "rotze Rotineque",
                        "corporation_id": 2002,
                        "corporation_name": "Eulenclub",
                    }
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

    @patch("ledger.api.character.ledger.CharacterAudit.objects.visible_eve_characters")
    def test_get_character_admin_no_visible(self, mock_visible_to):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/ledger/admin/"

        mock_visible_to.return_value = None

        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    def test_get_character_admin_exception(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/ledger/admin/"

        EveCorporationInfo.objects.get(corporation_id=2001).delete()

        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)



from unittest.mock import MagicMock, patch

from celery_once import AlreadyQueued
from corpstats.models import CorpMember as ExpectedCorpMember

from django.core.exceptions import ObjectDoesNotExist
from django.test import RequestFactory, TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.helpers import (
    _process_character,
    _process_missing_characters,
    get_alts_queryset,
    get_character,
    get_corp_models_and_string,
    get_main_and_alts_all,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.api.helpers"


class TestApiHelpers(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()

        cls.user, cls.char_owner = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
        )
        cls.user4, _ = create_user_from_evecharacter(
            1004,
        )
        cls.user5, _ = create_user_from_evecharacter(
            1005,
        )
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)

    def test_get_main_and_alts_all_char_in_chars(self):
        # given
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=1005, character_name="Gerthd", corpstats=corp_stats
        )
        chars = EveCharacter.objects.filter(
            corporation_id__in=[self.corp.corporation_id]
        )
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": [char]}
        excepted_data = mains
        # when
        data, _ = get_main_and_alts_all([self.corp.corporation_id])
        # then
        self.assertEqual(data, excepted_data)

    @patch(MODULE_PATH + ".EveCharacter.objects.get")
    def test_get_main_and_alts_all_process_corpmember(self, mock_eve_Character):
        # given
        mock_char = MagicMock()
        mock_char.character_id = 1005
        mock_char.character_name = "Gerthd"
        mock_eve_Character.return_value = mock_char

        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=9999, character_name="Test9999", corpstats=corp_stats
        )

        # when
        data, _ = get_main_and_alts_all([self.corp.corporation_id])
        # then
        mock_eve_Character.assert_called()
        self.assertIn("MagicMock", str(data.values()))

    @patch(MODULE_PATH + ".create_missing_character.apply_async")
    @patch(MODULE_PATH + ".EveCharacter.objects.get")
    def test_get_main_and_alts_all_process_corpmember_object_does_not_exist(
        self, mock_eve_Character, mock_apply_async
    ):
        # Setup mock EveCharacter instance to simulate ObjectDoesNotExist
        mock_eve_Character.return_value.side_effect = ObjectDoesNotExist

        # Setup the rest of your test environment as before
        mains = {}
        request = self.factory.get("/")
        request.user = self.user
        corp_stats = CorpStats.objects.create(
            token=Token.objects.get(user=self.user),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=9999, character_name="Test9999", corpstats=corp_stats
        )
        chars = EveCharacter.objects.filter(
            corporation_id__in=[self.corp.corporation_id]
        )
        for char in chars:
            mains[char.character_id] = {"main": char, "alts": [char]}

        # when
        data, _ = get_main_and_alts_all([self.corp.corporation_id])

    def test_get_main_character(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        add_character_to_user(
            self.user, EveCharacter.objects.get(character_id=1004), is_main=False
        )
        data = get_character(request, 1004)
        character = EveCharacter.objects.get(character_id=1004)
        # then
        self.assertEqual(data, (True, character))

    def test_get_main_character_no_permission(self):
        # given
        request = self.factory.get("/")
        request.user = self.user3
        # when
        data = get_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (False, character))

    @patch(MODULE_PATH + ".models.CharacterAudit.objects.visible_eve_characters")
    def test_main_char_not_in_account_chars(self, mock_visible):
        # given
        request = self.factory.get("/")
        mock_visible.return_value = []
        request.user = self.user
        # when
        data = get_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (False, character))

    def test_get_alts_queryset(self):
        # given
        main_char = EveCharacter.objects.get(character_id=1001)
        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)

    def test_get_alts_queryset_no_linked_characters(self):
        # given
        main_char = MagicMock()
        main_char.character_ownership.user.character_ownerships.all.return_value.values_list.side_effect = (
            ObjectDoesNotExist
        )

        existing_char = EveCharacter.objects.all().first()
        main_char.pk = existing_char.pk

        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)

    @patch(MODULE_PATH + ".app_settings.LEDGER_CORPSTATS_TWO", True)
    def test_get_corp_models_and_string(self):
        CorpMember = get_corp_models_and_string()
        self.assertIs(CorpMember, ExpectedCorpMember)

    def test_process_character(self):
        # Create mock main character
        main_char = EveCharacter(character_id=1002, corporation_id=2002)

        # Mock char with a main character
        char = EveCharacter(character_id=1001, corporation_id=2001)
        char.character_ownership = self.char_owner
        char.character_ownership.user.profile.main_character = main_char

        # Test main not in characters
        characters = {}
        chars_list = set()
        corporations = {}
        missing_chars = set()

        _process_character(char, characters, chars_list, corporations, missing_chars)

        # Test main in characters
        characters = {1002: {"main": main_char, "alts": []}}
        chars_list = set()
        corporations = {}
        missing_chars = set()

        _process_character(char, characters, chars_list, corporations, missing_chars)

        # Test Corporation exist
        characters = {}
        chars_list = set()
        corporations = [2001]
        missing_chars = set()

        _process_character(char, characters, chars_list, corporations, missing_chars)

        # Test Corporation not exist
        char = EveCharacter(character_id=1001, corporation_id=2001)
        characters = {}
        chars_list = set()
        corporations = {}
        missing_chars = set()

        _process_character(char, characters, chars_list, corporations, missing_chars)

        # Test Attribute Error
        error = EveCorporationInfo.objects.get(corporation_id=2001)
        characters = {}
        chars_list = set()
        corporations = [2001]
        missing_chars = set()

        _process_character(error, characters, chars_list, corporations, missing_chars)

        # Test Missing character
        char = EveCharacter(character_id=9999, corporation_id=2001)
        characters = {}
        chars_list = set()
        corporations = []
        missing_chars = set()

        _process_character(char, characters, chars_list, corporations, missing_chars)


class TestApiHelperTask(TestCase):
    @patch(MODULE_PATH + ".create_missing_character.apply_async")
    def test_already_queued_exception(self, mock_apply_async):
        # Given
        missing_chars = {1001}

        # Configure the mock to raise AlreadyQueued on the second call
        mock_apply_async.side_effect = [None, AlreadyQueued(countdown=60)]

        # When: First call should pass without raising an exception
        _process_missing_characters(missing_chars)

        # Then: Second call should raise AlreadyQueued, which is caught and handled
        _process_missing_characters(missing_chars)

        _process_missing_characters(None)

        # Assert apply_async was called twice
        self.assertEqual(mock_apply_async.call_count, 2)



noData = [
    {
        "ratting": [],
        "total": {
            "total_amount": 0,
            "total_amount_ess": 0,
            "total_amount_all": 0,
            "total_amount_mining": 0,
            "total_amount_others": 0,
            "total_amount_costs": 0,
        },
    }
]

CorpmonthlyMarch = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "alt_names": [],
                "total_amount": "200000.00",
                "total_amount_ess": "200000.00",
            }
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "200000.00",
            "total_amount_all": "400000.00",
            "total_amount_mining": 0,
            "total_amount_others": 0,
            "total_amount_costs": 0,
        },
    }
]


Corpyearly = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "alt_names": [],
                "total_amount": "200000.00",
                "total_amount_ess": "200000.00",
            }
        ],
        "total": {
            "total_amount": "200000.00",
            "total_amount_ess": "200000.00",
            "total_amount_all": "400000.00",
            "total_amount_mining": 0,
            "total_amount_others": 0,
            "total_amount_costs": 0,
        },
    }
]

Corpdatamany = [
    {
        "ratting": [
            {
                "main_id": 1001,
                "main_name": "Gneuten",
                "alt_names": [],
                "total_amount": "200000.00",
                "total_amount_ess": "200000.00",
            },
            {
                "main_id": 1010,
                "main_name": "Test1",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1019,
                "main_name": "Test10",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1020,
                "main_name": "Test11",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1021,
                "main_name": "Test12",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1011,
                "main_name": "Test2",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1012,
                "main_name": "Test3",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1013,
                "main_name": "Test4",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1014,
                "main_name": "Test5",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1015,
                "main_name": "Test6",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1016,
                "main_name": "Test7",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1017,
                "main_name": "Test8",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
            {
                "main_id": 1018,
                "main_name": "Test9",
                "alt_names": [],
                "total_amount": "100000.00",
                "total_amount_ess": "0.00",
            },
        ],
        "total": {
            "total_amount": "1400000.00",
            "total_amount_ess": "200000.00",
            "total_amount_all": "1600000.00",
            "total_amount_mining": 0,
            "total_amount_others": 0,
            "total_amount_costs": 0,
        },
    }
]



"""Admin models"""

from django.contrib import admin
from django.utils.html import format_html

from allianceauth.eveonline.evelinks import eveimageserver

from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit


@admin.register(CorporationAudit)
class CorporationAuditAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_pic",
        "_corporation__corporation_id",
        "_last_update_wallet",
    )

    list_display_links = (
        "_entity_pic",
        "_corporation__corporation_id",
    )

    list_select_related = ("corporation",)

    ordering = ["corporation__corporation_name"]

    search_fields = ["corporation__corporation_name", "corporation__corporation_id"]

    actions = [
        "delete_objects",
    ]

    @admin.display(description="")
    def _entity_pic(self, obj: CorporationAudit):
        eve_id = obj.corporation.corporation_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("corporation", eve_id, 32),
        )

    @admin.display(description="Corporation ID", ordering="corporation__corporation_id")
    def _corporation__corporation_id(self, obj: CorporationAudit):
        return obj.corporation.corporation_id

    @admin.display(description="Last Update Wallet", ordering="last_update_wallet")
    def _last_update_wallet(self, obj: CorporationAudit):
        return obj.last_update_wallet

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False


@admin.register(CharacterAudit)
class CharacterAuditAdmin(admin.ModelAdmin):
    list_display = (
        "_entity_pic",
        "_character__character_name",
        "_last_update_wallet",
        "_last_update_mining",
    )

    list_display_links = (
        "_entity_pic",
        "_character__character_name",
    )

    list_select_related = ("character",)

    ordering = ["character__character_name"]

    search_fields = ["character__character_name"]

    actions = [
        "delete_objects",
    ]

    @admin.display(description="")
    def _entity_pic(self, obj: CharacterAudit):
        eve_id = obj.character.character_id
        return format_html(
            '<img src="{}" class="img-circle">',
            eveimageserver._eve_entity_image_url("character", eve_id, 32),
        )

    @admin.display(description="Character Name", ordering="character__character_name")
    def _character__character_name(self, obj: CharacterAudit):
        return obj.character.character_name

    @admin.display(description="Last Update Wallet", ordering="last_update_wallet")
    def _last_update_wallet(self, obj: CharacterAudit):
        return obj.last_update_wallet

    @admin.display(description="Last Update Mining", ordering="last_update_mining")
    def _last_update_mining(self, obj: CharacterAudit):
        return obj.last_update_mining

    # pylint: disable=unused-argument
    def has_add_permission(self, request):
        return False

    # pylint: disable=unused-argument
    def has_change_permission(self, request, obj=None):
        return False



"""PvE Views"""

from datetime import datetime

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

from ledger.hooks import get_extension_logger

# Ledger
from ledger.view_helpers.core import add_info_to_context

logger = get_extension_logger(__name__)


@login_required
@permission_required("ledger.basic_access")
def character_ledger(request, character_pk):
    """
    Character Ledger
    """
    current_year = datetime.now().year
    years = [current_year - i for i in range(6)]

    context = {
        "years": years,
        "character_pk": character_pk,
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/charledger/character_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access")
def character_admin(request):
    """
    Character Admin
    """

    context = {}
    context = add_info_to_context(request, context)

    return render(
        request, "ledger/charledger/admin/character_admin.html", context=context
    )



from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from ledger.api import character, corporation
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)

api = NinjaAPI(
    title="Geuthur API",
    version="0.1.0",
    urls_namespace="ledger:new_api",
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the character endpoints
character.setup(api)

# Add the corporation endpoints
corporation.setup(api)



from datetime import timedelta
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from ledger.models.general import EveEntity

MODULE_PATH = "ledger.models.general"


class TestGeneralModel(TestCase):
    def setUp(self):
        self.eveentity = EveEntity()
        self.eveentity.name = "Test"
        self.eveentity.eve_id = 123
        self.eveentity.category = "Test Category"
        self.eveentity.last_update = timezone.now() - timedelta(days=8)

    def test_str(self):
        self.assertEqual(str(self.eveentity), "Test")

    def test_repr(self):
        self.assertEqual(
            repr(self.eveentity),
            "EveEntity(id=123, category='Test Category', name='Test')",
        )

    def test_is_alliance(self):
        self.eveentity.category = self.eveentity.CATEGORY_ALLIANCE
        self.assertTrue(self.eveentity.is_alliance)

    def test_is_corporation(self):
        self.eveentity.category = self.eveentity.CATEGORY_CORPORATION
        self.assertTrue(self.eveentity.is_corporation)

    def test_is_character(self):
        self.eveentity.category = self.eveentity.CATEGORY_CHARACTER
        self.assertTrue(self.eveentity.is_character)

    @patch(
        MODULE_PATH + ".EveAllianceInfo.generic_logo_url"
    )  # replace with your actual module path
    def test_icon_url_alliance(self, mock_generic_logo_url):
        self.eveentity.category = self.eveentity.CATEGORY_ALLIANCE
        mock_generic_logo_url.return_value = "Test URL"
        self.assertEqual(self.eveentity.icon_url(), "Test URL")

    @patch(
        MODULE_PATH + ".EveCorporationInfo.generic_logo_url"
    )  # replace with your actual module path
    def test_icon_url_corporation(self, mock_generic_logo_url):
        self.eveentity.category = self.eveentity.CATEGORY_CORPORATION
        mock_generic_logo_url.return_value = "Test URL"
        self.assertEqual(self.eveentity.icon_url(), "Test URL")

    @patch(
        MODULE_PATH + ".EveCharacter.generic_portrait_url"
    )  # replace with your actual module path
    def test_icon_url_character(self, mock_generic_portrait_url):
        self.eveentity.category = self.eveentity.CATEGORY_CHARACTER
        mock_generic_portrait_url.return_value = "Test URL"
        self.assertEqual(self.eveentity.icon_url(), "Test URL")

    def test_icon_url_not_implemented(self):
        self.eveentity.category = "Invalid Category"
        with self.assertRaises(NotImplementedError):
            self.eveentity.icon_url()

    def test_needs_update(self):
        self.assertTrue(self.eveentity.needs_update())



from unittest.mock import MagicMock, Mock, patch

from memberaudit.models import Character

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.models.characteraudit import CharacterAudit
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_memberaudit import load_memberaudit
from ledger.views.character.char_audit import add_char, fetch_memberaudit

MODULE_PATH = "ledger.views.character.char_audit"


class CharAuditTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_memberaudit()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.char_audit_admin_manager",
            ],
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    def test_add_char(self, mock_update_character, _):
        self.client.force_login(self.user)
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        request = self.factory.get(reverse("ledger:ledger_add_char"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_char.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_update_character.apply_async.called)
        self.assertTrue(
            CharacterAudit.objects.get(character=self.character_ownership.character)
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    def test_fetch_memberaudit_updates(self, mock_update_character, mock_messages):
        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = fetch_memberaudit(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_messages.info.called)
        mock_messages.info.assert_called_with(
            request, "3 Char(s) successfully added/updated to Ledger"
        )
        self.assertTrue(mock_update_character.apply_async.called)
        self.assertTrue(
            CharacterAudit.objects.get(character=self.character_ownership.character)
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch(MODULE_PATH + ".CharacterAudit.objects.get_or_create")
    def test_fetch_memberaudit_no_updates(
        self, mock_char_audit, mock_update_character, mock_messages
    ):
        # Set last_update_wallet to now to prevent update
        char_audit = Mock()
        char_audit.last_update_wallet = timezone.now()
        mock_char_audit.return_value = (char_audit, False)

        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = fetch_memberaudit(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_messages.info.called)
        mock_messages.info.assert_called_with(request, "No Updates initialized.")
        self.assertFalse(mock_update_character.apply_async.called)

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch(MODULE_PATH + ".CharacterAudit.objects.get_or_create")
    def test_fetch_memberaudit_no_char_audit(
        self, mock_char_audit, mock_update_character, mock_messages
    ):
        # Set last_update_wallet to now to prevent update
        mock_char_audit.return_value = (None, False)

        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = fetch_memberaudit(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_messages.info.called)
        mock_messages.info.assert_called_with(request, "No Updates initialized.")
        self.assertFalse(mock_update_character.apply_async.called)

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch("memberaudit.models.Character.objects.filter")
    def test_fetch_memberaudit_general_error(
        self, mock_filter, mock_update_character, mock_messages
    ):
        mock_filter.side_effect = Exception
        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = fetch_memberaudit(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.error.called)
        mock_messages.error.assert_called_with(
            request, "An error occurred: Please inform your Admin."
        )
        self.assertFalse(mock_update_character.apply_async.called)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch("memberaudit.models.Character.objects.filter")
    def test_fetch_memberaudit_import_error(
        self, mock_filter, mock_update_character, mock_messages
    ):
        mock_filter.side_effect = ImportError
        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # when
        response = fetch_memberaudit(request)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertTrue(mock_messages.error.called)
        mock_messages.error.assert_called_with(
            request,
            "The 'memberaudit' app is not installed. Please make sure it is installed.",
        )
        self.assertFalse(mock_update_character.apply_async.called)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch("memberaudit.models.Character.objects.filter")
    @patch(MODULE_PATH + ".CharacterAudit.objects.values_list")
    def test_char_in_ledger_chars_ids(
        self, mock_values_list, mock_filter, mock_update_character, mock_messages
    ):
        mock_values_list.return_value = [9999]
        mock_char = MagicMock()
        mock_char.eve_character.character_id = 9999
        mock_filter.return_value = [mock_char]

        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        response = fetch_memberaudit(request)

        self.assertEqual(response.status_code, 302)
        self.assertFalse(mock_update_character.apply_async.called)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch("memberaudit.models.Character.objects.filter")
    @patch(MODULE_PATH + ".CharacterAudit.objects.get_or_create")
    def test_eve_character_does_not_exist(
        self, mock_get_or_create, mock_filter, mock_update_character, mock_messages
    ):
        mock_get_or_create.side_effect = EveCharacter.DoesNotExist
        mock_char = MagicMock()
        mock_char.eve_character.character_id = 9999
        mock_filter.return_value = [mock_char]

        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        response = fetch_memberaudit(request)

        self.assertEqual(response.status_code, 302)
        self.assertFalse(mock_update_character.apply_async.called)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_character")
    @patch("memberaudit.models.Character.objects.filter")
    @patch(MODULE_PATH + ".CharacterAudit.objects.get_or_create")
    def test_eve_character_attribute_error(
        self, mock_get_or_create, mock_filter, mock_update_character, mock_messages
    ):
        mock_get_or_create.side_effect = AttributeError
        mock_char = MagicMock()
        mock_char.eve_character.character_id = 9999
        mock_filter.return_value = [mock_char]

        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_fetch_memberaudit"))
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)

        response = fetch_memberaudit(request)

        self.assertEqual(response.status_code, 302)
        self.assertFalse(mock_update_character.apply_async.called)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))



"""
Core View Helper
"""

from datetime import datetime, timedelta
from decimal import Decimal

from django.core.cache import cache
from django.db.models import Q

from allianceauth.authentication.models import UserProfile

from ledger.app_settings import STORAGE_BASE_KEY
from ledger.hooks import get_extension_logger
from ledger.models.events import Events

logger = get_extension_logger(__name__)


# pylint: disable=unused-argument
def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    theme = None
    try:
        user = UserProfile.objects.get(id=request.user.id)
        theme = user.theme
    except UserProfile.DoesNotExist:
        pass

    new_context = {
        **{"theme": theme},
        **context,
    }
    return new_context


def ledger_cache_timeout():
    """
    Calculate time left to next hour

    Example:
    1 Hour
        10:15 -> 45 minutes left
        10:45 -> 15 minutes left
    """
    now = datetime.now()
    delta = timedelta(minutes=60, hours=0, seconds=0)
    next_time = (now + delta).replace(minute=0, second=0, microsecond=0)
    timeout = next_time - now
    return timeout.total_seconds()


def calculate_ess_stolen_amount(bounty, ess):
    try:
        total_ess_gain = ess / bounty
        total_ess_gain = total_ess_gain * 100

        total_ess_gain_diff = Decimal(66.667) - int(total_ess_gain)

        stolen = ess * (total_ess_gain_diff / 100)
        # If the difference is less than 0.9 or negative, no ESS was stolen
        if abs(total_ess_gain_diff) < Decimal("0.9") or total_ess_gain_diff < 0:
            stolen = 0
    # pylint: disable=broad-except
    except Exception:
        stolen = 0
    return round(stolen)


def calculate_ess_stolen(amounts):
    try:
        amounts["stolen"]["total_amount"] = calculate_ess_stolen_amount(
            amounts["bounty"]["total_amount"], amounts["ess"]["total_amount"]
        )
        amounts["stolen"]["total_amount_day"] = calculate_ess_stolen_amount(
            amounts["bounty"]["total_amount_day"], amounts["ess"]["total_amount_day"]
        )
    except KeyError:
        pass
    return amounts


def _storage_key(identifier: str) -> str:
    return STORAGE_BASE_KEY + str(identifier)


def get_cache_stale(key_name):
    data = cache.get(key=_storage_key(key_name))
    if not data:
        return False
    return data


def set_cache(data, key_name: str, time: int):
    cache.set(
        key=_storage_key(key_name),
        value=data,
        timeout=time,
    )


def set_cache_hourly(data, key_name: str):
    """Reset hourly"""
    cache.set(
        key=_storage_key(key_name),
        value=data,
        timeout=ledger_cache_timeout(),
    )


def delete_cache(key_name: str):
    cache.delete(_storage_key(key_name))


def events_filter(entries):
    """
    Filter out all Entries that are in the time of the Event
    """
    # Events to Filter out
    events = Events.objects.all()

    q_objects = []

    # Durchlaufen Sie jedes Event und erstellen Sie das entsprechende Q-Objekt für den Datumsbereich
    for event in events:
        if not event.char_ledger:
            continue
        q_objects.append(Q(date__range=(event.date_start, event.date_end)))

    # Kombinieren Sie alle Q-Objekte mit einer ODER-Verknüpfung
    if q_objects:
        combined_q_object = q_objects[0]
        for q_object in q_objects[1:]:
            combined_q_object |= q_object

        entries = entries.exclude(combined_q_object)
    return entries



from unittest.mock import MagicMock, Mock, patch

from django.test import TestCase

from ledger.task_helpers.etag_helpers import (
    MAX_ETAG_LIFE,
    HTTPNotModified,
    NotModifiedError,
    del_etag_header,
    etag_results,
    get_etag_header,
    get_etag_key,
    handle_etag_headers,
    handle_page_results,
    inject_etag_header,
    rem_etag_header,
    set_etag_header,
    stringify_params,
)

MODULE_PATH = "ledger.task_helpers.etag_helpers"


class TestEtagHelpers(TestCase):
    def setUp(self):
        self.mock_operation = Mock()
        self.mock_operation._cache_key.return_value = "cache_key"
        self.mock_operation.operation.operation_id = "operation_id"
        self.mock_operation.future.request.headers = {}
        self.mock_operation.future.request.params = {
            "param1": "value1",
            "param2": "value2",
        }

    def test_get_etag_key(self):
        result = get_etag_key(self.mock_operation)
        self.assertEqual(result, "etag-cache_key")

    @patch(MODULE_PATH + ".cache.get")
    def test_get_etag_header(self, mock_cache_get):
        mock_cache_get.return_value = "ABCD"
        result = get_etag_header(self.mock_operation)
        self.assertEqual(result, "ABCD")

    @patch(MODULE_PATH + ".cache.delete")
    def test_del_etag_header(self, mock_cache_delete):
        del_etag_header(self.mock_operation)
        mock_cache_delete.assert_called_once_with("etag-cache_key", False)

    @patch(MODULE_PATH + ".get_etag_header")
    def test_inject_etag_header(self, mock_get_etag_header):
        mock_get_etag_header.return_value = "ABCD"
        inject_etag_header(self.mock_operation)
        self.assertEqual(
            self.mock_operation.future.request.headers["If-None-Match"], "ABCD"
        )

    def test_rem_etag_header(self):
        self.mock_operation.future.request.headers["If-None-Match"] = "ABCD"
        result = rem_etag_header(self.mock_operation)
        self.assertNotIn("If-None-Match", self.mock_operation.future.request.headers)
        self.assertTrue(result)

        self.mock_operation.future.request.headers = {}
        result = rem_etag_header(self.mock_operation)
        self.assertFalse(result)

    @patch(MODULE_PATH + ".get_etag_key")
    @patch(MODULE_PATH + ".cache.set")
    def test_set_etag_header(self, mock_cache_set, mock_get_etag_key):
        mock_get_etag_key.return_value = "etag_key"
        headers = Mock()
        headers.headers.get.return_value = "ABCD"
        result = set_etag_header(self.mock_operation, headers)
        mock_cache_set.assert_called_once_with("etag_key", "ABCD", MAX_ETAG_LIFE)
        self.assertTrue(result)

        mock_get_etag_key.return_value = None
        headers.headers.get.return_value = None
        result = set_etag_header(self.mock_operation, headers)
        self.assertFalse(result)

    def test_stringify_params(self):
        result = stringify_params(self.mock_operation)
        self.assertEqual(result, "param1: value1, param2: value2")


class TestEtagReults(TestCase):
    def setUp(self):
        self.operation = MagicMock()
        self.headers = MagicMock()  # Use MagicMock instead of Mock
        self.headers.headers = {"ETag": "ABCD", "X-Pages": "2"}

        self.mock_operation = MagicMock()
        self.mock_operation._cache_key.return_value = "cache_key"
        self.mock_operation.operation.operation_id = "operation_id"
        self.mock_operation.future.request.headers = {"If-None-Match": "ABCD"}
        self.mock_operation.future.request.params = {
            "param1": "value1",
            "param2": "value2",
        }

    @patch("django_redis.cache.RedisCache.set")
    def test_handle_page_results(self, _):
        self.operation.result.return_value = ([], self.headers)
        results, current_page, total_pages = handle_page_results(
            self.operation, 1, 2, False, False
        )
        self.assertEqual(results, [])
        self.assertEqual(current_page, 3)
        self.assertEqual(total_pages, 2)

    @patch(MODULE_PATH + ".handle_etag_headers")
    @patch("django_redis.cache.RedisCache.set")
    def test_handle_page_results_notmodified(self, _, mock_handle_etag_headers):
        # given
        self.operation.result.return_value = ([], self.headers)
        mock_error = NotModifiedError()
        mock_error.response = Mock()
        mock_error.response.headers = self.headers
        mock_handle_etag_headers.side_effect = mock_error

        # when
        results, current_page, total_pages = handle_page_results(
            self.operation, 1, 2, False, False
        )

        # then
        mock_handle_etag_headers.assert_called()
        self.assertEqual(current_page, 3)
        self.assertEqual(total_pages, 2)
        self.assertEqual(results, [])

    @patch(MODULE_PATH + ".handle_etag_headers")
    def test_handle_page_results_http_not_modified(self, mock_handle_etag_headers):
        self.operation.result.return_value = ([], self.headers)
        mock_handle_etag_headers.side_effect = HTTPNotModified(self.headers)
        # when
        results, current_page, total_pages = handle_page_results(
            self.operation, 1, 2, False, False
        )
        # then
        self.assertEqual(current_page, 3)
        self.assertEqual(total_pages, 2)

    @patch(MODULE_PATH + ".handle_etag_headers")
    def test_handle_page_results_http_not_modified_etags_incomplete(
        self, mock_handle_etag_headers
    ):
        self.operation.result.return_value = ([], self.headers)
        mock_handle_etag_headers.side_effect = HTTPNotModified(self.headers)

        # when
        results, current_page, total_pages = handle_page_results(
            self.operation, 1, 2, True, False
        )

        # then
        self.assertEqual(current_page, 3)
        self.assertEqual(total_pages, 2)
        self.assertEqual(results, [])

    @patch(MODULE_PATH + ".get_etag_header")
    @patch(MODULE_PATH + ".del_etag_header")
    @patch(MODULE_PATH + ".set_etag_header")
    @patch(MODULE_PATH + ".stringify_params")
    def test_handle_etag_headers(
        self,
        mock_stringify_params,
        mock_set_etag_header,
        mock_del_etag_header,
        mock_get_etag_header,
    ):
        # given
        mock_operation = MagicMock()
        mock_operation.operation.operation_id = "test_id"
        mock_headers = MagicMock()
        mock_headers.headers.get.return_value = "test_etag"
        mock_get_etag_header.return_value = "test_etag"
        mock_stringify_params.return_value = "test_params"

        # when
        with self.assertRaises(NotModifiedError):
            handle_etag_headers(mock_operation, mock_headers, False, False)

        # then
        mock_get_etag_header.assert_called_once_with(mock_operation)
        mock_set_etag_header.assert_not_called()
        mock_del_etag_header.assert_not_called()

    @patch(MODULE_PATH + ".handle_etag_headers")
    def test_etag_results(
        self,
        _,
    ):
        # given
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = "valid_token"

        mock_results = ["result1", "result2"]
        mock_headers = MagicMock()

        self.mock_operation.result.return_value = (mock_results, mock_headers)

        # when
        results = etag_results(self.mock_operation, mock_token)

        # then
        self.assertEqual(results, mock_results)
        mock_token.valid_access_token.assert_called_once()

    def test_etag_results_notoken(
        self,
    ):
        # given
        mock_token = False

        mock_results = ["result1", "result2"]
        mock_headers = MagicMock()

        self.mock_operation.result.return_value = (mock_results, mock_headers)

        # when
        results = etag_results(self.mock_operation, mock_token, force_refresh=True)

        # then
        self.assertEqual(results, mock_results)

    @patch(MODULE_PATH + ".handle_page_results")
    @patch(MODULE_PATH + ".set_etag_header")
    def test_etag_results_HTTPNotModified(
        self, mock_set_etag_header, mock_handle_page_results
    ):
        # given
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = "valid_token"

        mock_results = ["result1", "result2"]

        mock_response = MagicMock()

        self.mock_operation.result.side_effect = HTTPNotModified(mock_response)
        self.mock_operation.operation.params = {}
        mock_handle_page_results.return_value = (mock_results, 1, 1)

        # when
        with self.assertRaises(NotModifiedError):
            etag_results(self.mock_operation, mock_token)

        # then
        mock_token.valid_access_token.assert_called_once()
        mock_handle_page_results.assert_not_called()
        mock_set_etag_header.assert_called_once()

    @patch(MODULE_PATH + ".handle_page_results")
    def test_etag_results_with_page(self, mock_handle_page_results):
        # given
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = "valid_token"

        mock_results = ["result1", "result2"]
        mock_response = MagicMock()

        self.mock_operation.return_value = mock_response
        self.mock_operation.operation.params = {"page": 1}
        mock_handle_page_results.return_value = (mock_results, 1, 1)
        # when
        results = etag_results(self.mock_operation, mock_token)

        # then
        mock_token.valid_access_token.assert_called_once()
        mock_handle_page_results.assert_called_once_with(
            self.mock_operation, 1, 1, False, False
        )
        self.assertEqual(results, mock_results)
        self.assertEqual(
            self.mock_operation.future.request.headers["If-None-Match"], "ABCD"
        )

    @patch(MODULE_PATH + ".handle_page_results")
    def test_etag_results_with_page_notmodified(self, mock_handle_page_results):
        # given
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = "valid_token"

        mock_results = ["result1", "result2"]
        mock_response = MagicMock()

        self.mock_operation.result.side_effect = HTTPNotModified(mock_response)
        self.mock_operation.operation.params = {"page": 1}
        mock_handle_page_results.return_value = (mock_results, 1, 1)
        # when
        results = etag_results(self.mock_operation, mock_token)

        # then
        mock_token.valid_access_token.assert_called_once()
        mock_handle_page_results.assert_called_once_with(
            self.mock_operation, 1, 1, False, False
        )
        self.assertEqual(results, mock_results)



from typing import List

from ninja import NinjaAPI

from ledger.api import schema
from ledger.api.helpers import get_alts_queryset, get_character
from ledger.api.managers.character_manager import CharacterProcess
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterAudit

logger = get_extension_logger(__name__)


# pylint: disable=duplicate-code
class LedgerApiEndpoints:
    tags = ["CharacerLedger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "account/{character_id}/ledger/year/{year}/month/{month}/",
            response={200: List[schema.Ledger], 403: str},
            tags=self.tags,
        )
        def get_character_ledger(request, character_id: int, year: int, month: int):
            perm, main = get_character(request, character_id)

            if not perm:
                return 403, "Permission Denied"

            # Create the Ledger
            if character_id == 0:
                characters = get_alts_queryset(main)
            else:
                characters = [main]

            ledger = CharacterProcess(characters, year, month)
            output = ledger.generate_ledger()

            return output

        @api.get(
            "account/{character_id}/billboard/year/{year}/month/{month}/",
            response={200: List[schema.Billboard], 403: str},
            tags=self.tags,
        )
        def get_billboard_ledger(request, character_id: int, year: int, month: int):
            perm, main = get_character(request, character_id)

            if not perm:
                return 403, "Permission Denied"

            # Create the Ledger
            if character_id == 0:
                characters = get_alts_queryset(main)
            else:
                characters = [main]

            ledger = CharacterProcess(characters, year, month)
            output = ledger.generate_billboard()

            return output

        @api.get(
            "account/ledger/admin/",
            response={200: List[schema.CharacterAdmin], 403: str},
            tags=self.tags,
        )
        def get_character_admin(request):
            characters = CharacterAudit.objects.visible_eve_characters(request.user)

            if not characters:
                return 403, "Permission Denied"

            character_dict = {}

            for character in characters:
                # pylint: disable=broad-exception-caught
                try:
                    character_dict[character.character_id] = {
                        "character_id": character.character_id,
                        "character_name": character.character_name,
                        "corporation_id": character.corporation_id,
                        "corporation_name": character.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"character": character_dict})

            return output



"""
Events Model
"""

from django.db import models
from django.utils.translation import gettext_lazy as _


# TODO Make it Corp Specific
class Events(models.Model):
    """Event Tracker for Voices of War."""

    id = models.BigAutoField(primary_key=True)
    title = models.CharField(max_length=100, null=True)
    date_start = models.DateTimeField(null=True, blank=True)
    date_end = models.DateTimeField(null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    char_ledger = models.BooleanField(null=True, default=False)
    location = models.CharField(max_length=100, null=True, blank=True)

    class Meta:
        verbose_name = _("event")
        verbose_name_plural = _("events")
        default_permissions = ()
        permissions = (
            (
                "event_admin_access",
                "Can access Events Tools",
            ),
        )

    def __str__(self):
        return f"Event {self.id}"



from datetime import datetime
from typing import Dict, List, Optional

from ninja import Schema

from ledger.api.managers.billboard_manager import BillboardDict


class Message(Schema):
    message: str


class Character(Schema):
    character_name: str
    character_id: int
    corporation_id: int
    corporation_name: str
    alliance_id: Optional[int] = None
    alliance_name: Optional[str] = None


class Corporation(Schema):
    corporation_id: int
    corporation_name: str
    alliance_id: Optional[int] = None
    alliance_name: Optional[str] = None


class EveName(Schema):
    id: int
    name: str
    cat: Optional[str] = None


class CharacterWalletEvent(Schema):
    character: Optional[Character] = None
    id: Optional[int] = None
    date: Optional[datetime] = None
    first_party: Optional[EveName] = None
    second_party: Optional[EveName] = None
    ref_type: Optional[str] = None
    balance: Optional[float] = None
    amount: Optional[float] = None
    reason: Optional[str] = None


class CharacterLedger(Schema):
    ratting: Optional[List] = None
    total: Optional[Dict] = None
    billboard: Optional[BillboardDict] = None


class CharacterLedgerTemplate(Schema):
    character: Optional[str] = None


class CharacterAdmin(Schema):
    character: Optional[dict] = None


class CorporationAdmin(Schema):
    corporation: Optional[dict] = None


class CorporationWalletEvent(Schema):
    division: str
    id: int
    date: datetime
    first_party: EveName
    second_party: EveName
    ref_type: str
    balance: float
    amount: float
    reason: Optional[str] = None


class Ledger(Schema):
    ratting: Optional[List] = None
    total: Optional[Dict] = None


class Billboard(Schema):
    billboard: Optional[BillboardDict] = None



"""
Corporation Events
"""

# Django
from django import forms
from django.contrib.auth.decorators import login_required, permission_required
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils import timezone
from django.utils.html import escape

from ledger.hooks import get_extension_logger

# Voices of War
from ledger.models.events import Events

logger = get_extension_logger(__name__)


class EventForm(forms.ModelForm):
    class Meta:
        model = Events
        fields = [
            "title",
            "date_start",
            "date_end",
            "location",
            "description",
            "char_ledger",
        ]
        widgets = {
            "date_start": forms.DateTimeInput(attrs={"type": "datetime-local"}),
            "date_end": forms.DateTimeInput(attrs={"type": "datetime-local"}),
            "description": forms.Textarea(attrs={"rows": 6}),
            "char_ledger": forms.CheckboxInput(attrs={"class": "form-check-input"}),
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["title"].required = True
        self.fields["date_start"].required = True
        self.fields["date_end"].required = True


@login_required
@permission_required("ledger.basic_access")
def events_index(request):
    current_datetime = timezone.now()
    expired_events = Events.objects.filter(date_end__lt=current_datetime)
    future_events = Events.objects.filter(date_end__gte=current_datetime)

    return render(
        request,
        "ledger/events/index.html",
        {"expired_events": expired_events, "future_events": future_events},
    )


@login_required
@permission_required("ledger.event_admin_access")
def events_admin(request):
    events = Events.objects.all()
    return render(request, "ledger/events/manage_events.html", {"events": events})


@login_required
@permission_required("ledger.event_admin_access")
def create_event(request):
    if request.method == "POST":
        form = EventForm(request.POST)
        if form.is_valid():
            form.save()
            return redirect("ledger:event_admin")
    else:
        form = EventForm()
    return render(request, "ledger/events/create_event.html", {"form": form})


@login_required
@permission_required("ledger.event_admin_access")
def edit_event(request, event_id):
    event = get_object_or_404(Events, pk=event_id)
    if request.method == "POST":
        form = EventForm(request.POST, instance=event)
        if form.is_valid():
            form.save()
            return redirect("ledger:event_admin")
    else:
        event.date_start = (
            event.date_start.strftime("%Y-%m-%dT%H:%M") if event.date_start else None
        )
        event.date_end = (
            event.date_end.strftime("%Y-%m-%dT%H:%M") if event.date_end else None
        )

        form = EventForm(instance=event)
    return render(
        request, "ledger/events/edit_event.html", {"form": form, "event": event}
    )


@login_required
@permission_required("ledger.event_admin_access")
def delete_event(request, event_id):
    event = get_object_or_404(Events, pk=event_id)
    if request.method == "POST":
        event.delete()
        return redirect("ledger:event_admin")
    return render(request, "ledger/events/delete_event.html", {"event": event})


# pylint: disable=unused-argument
def load_events(request):
    events = Events.objects.all()
    formatted_events = []
    for event in events:
        formatted_description = escape(event.description).replace("\n", "<br>")
        formatted_events.append(
            {
                "id": event.id,
                "title": event.title,
                "date_start": event.date_start,
                "date_end": event.date_end,
                "description": formatted_description,
                "char_ledger": event.char_ledger,
                "location": event.location,
                "upcoming": event.date_end >= timezone.now(),
            }
        )

    return JsonResponse(formatted_events, safe=False)



from typing import List

from ninja import NinjaAPI

from ledger.api import schema
from ledger.api.helpers import get_corporation, get_main_and_alts_corporations
from ledger.api.managers.corporation_manager import CorporationProcess
from ledger.hooks import get_extension_logger
from ledger.models import CorporationAudit

logger = get_extension_logger(__name__)


class LedgerApiEndpoints:
    tags = ["CorporationLedger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/ledger/year/{year}/month/{month}/",
            response={200: List[schema.Ledger], 403: str},
            tags=self.tags,
        )
        def get_corporation_ledger(request, corporation_id: int, year: int, month: int):
            response, _ = get_corporation(request, corporation_id)

            if not response:
                return 403, "Permission Denied"

            # pylint: disable=duplicate-code
            if corporation_id == 0:
                corporations = get_main_and_alts_corporations(request)
            else:
                corporations = [corporation_id]

            ledger = CorporationProcess(year, month, corporations)
            output = ledger.generate_ledger()

            return output

        @api.get(
            "corporation/{corporation_id}/billboard/year/{year}/month/{month}/",
            response={200: List[schema.Billboard], 403: str},
            tags=self.tags,
        )
        def get_billboard_ledger(request, corporation_id: int, year: int, month: int):
            response, _ = get_corporation(request, corporation_id)

            if not response:
                return 403, "Permission Denied"

            # pylint: disable=duplicate-code
            if corporation_id == 0:
                corporations = get_main_and_alts_corporations(request)
            else:
                corporations = [corporation_id]

            ledger = CorporationProcess(year, month, corporations)
            output = ledger.generate_billboard(corporations)

            return output

        @api.get(
            "corporation/ledger/admin/",
            response={200: List[schema.CorporationAdmin], 403: str},
            tags=self.tags,
        )
        def get_corporation_admin(request):
            corporations = CorporationAudit.objects.visible_to(request.user)

            if not corporations:
                return 403, "Permission Denied"

            corporation_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    corporation_dict[corporation.corporation.corporation_id] = {
                        "corporation_id": corporation.corporation.corporation_id,
                        "corporation_name": corporation.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"corporation": corporation_dict})

            return output



from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.corpaudit_manager"


class CorpAuditQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

    def test_visible_to_superuser(self):
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
        )
        self.user.is_superuser = True
        self.user.save()

        # Call the visible_to method with the superuser
        result = CorporationAudit.objects.visible_to(self.user)

        # Check that all CorporationAudit objects are returned
        expected_result = CorporationAudit.objects.all()
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_corp_audit_admin(self):
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.corp_audit_admin_manager",
            ],
        )

        expected_result = CorporationAudit.objects.all()

        result = CorporationAudit.objects.visible_to(self.user)
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_corp_manager_access(self):
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.corp_audit_manager",
            ],
        )

        expected_result = CorporationAudit.objects.filter(id=1)

        result = CorporationAudit.objects.visible_to(self.user)
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_error(self):
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        self.user.profile.main_character = None

        result = (
            CorporationAudit.objects.visible_to(self.user)
            .values_list("corporation_id", flat=True)
            .count()
        )

        self.assertEqual(result, 0)

    def test_visible_to_no_access(self):
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
        )

        result = CorporationAudit.objects.visible_to(self.user).count()

        self.assertEqual(result, 1)



"""
General Model
"""

import datetime

from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)

from ledger.hooks import get_extension_logger
from ledger.managers.general_manager import EveEntityManager

logger = get_extension_logger(__name__)

# Permission Manager


class General(models.Model):
    """A model defining commonly used properties and methods for Voices of War."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app, Ledger."),
            ("admin_access", "Has access to all Administration tools"),
        )


# EvE Entity Model - Store all Chars, Corps, Allys
class EveEntity(models.Model):
    """An Eve entity like a corporation or a character"""

    CATEGORY_ALLIANCE = "alliance"
    CATEGORY_CHARACTER = "character"
    CATEGORY_CORPORATION = "corporation"

    CATEGORY_CHOICES = (
        (CATEGORY_ALLIANCE, "Alliance"),
        (CATEGORY_CORPORATION, "Corporation"),
        (CATEGORY_CHARACTER, "Character"),
    )

    eve_id = models.IntegerField(
        primary_key=True, validators=[MinValueValidator(0)], verbose_name=_("id")
    )
    category = models.CharField(
        max_length=32, choices=CATEGORY_CHOICES, verbose_name=_("category")
    )
    name = models.CharField(max_length=254, verbose_name=_("name"))

    # optionals for character/corp
    corporation = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="corp",
    )
    alliance = models.ForeignKey(
        "EveEntity",
        on_delete=models.SET_NULL,
        null=True,
        default=None,
        related_name="alli",
    )
    last_update = models.DateTimeField(auto_now=True)

    objects = EveEntityManager()

    def __str__(self) -> str:
        return str(self.name)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.eve_id}, category='{self.category}', "
            f"name='{self.name}')"
        )

    @property
    def is_alliance(self) -> bool:
        """Return True if entity is an alliance, else False."""
        return self.category == self.CATEGORY_ALLIANCE

    @property
    def is_corporation(self) -> bool:
        """Return True if entity is an corporation, else False."""
        return self.category == self.CATEGORY_CORPORATION

    @property
    def is_character(self) -> bool:
        """Return True if entity is a character, else False."""
        return self.category == self.CATEGORY_CHARACTER

    def icon_url(self, size=128) -> str:
        """Url to an icon image for this organization."""
        if self.category == self.CATEGORY_ALLIANCE:
            return EveAllianceInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CORPORATION:
            return EveCorporationInfo.generic_logo_url(self.eve_id, size=size)

        if self.category == self.CATEGORY_CHARACTER:
            return EveCharacter.generic_portrait_url(self.eve_id, size=size)

        raise NotImplementedError(
            f"Avatar URL not implemented for category {self.category}"
        )

    def needs_update(self):
        return self.last_update + datetime.timedelta(days=7) < timezone.now()

    class Meta:
        default_permissions = ()



"""App URLs"""

from django.urls import path, re_path

from ledger.api import api
from ledger.views.character.char_audit import add_char, fetch_memberaudit
from ledger.views.character.character_ledger import character_admin, character_ledger
from ledger.views.corporation.corp_audit import add_corp
from ledger.views.corporation.corp_events import (
    create_event,
    delete_event,
    edit_event,
    events_admin,
    events_index,
    load_events,
)
from ledger.views.corporation.corporation_ledger import (
    corporation_admin,
    corporation_ledger,
)

# AA Example App
from ledger.views.pve import ledger_index

app_name: str = "ledger"

urlpatterns = [
    path("", ledger_index, name="index"),
    # -- Character Audit
    path("char/add/", add_char, name="ledger_add_char"),
    re_path(
        r"^char/fetch_memberaudit/", fetch_memberaudit, name="ledger_fetch_memberaudit"
    ),
    # -- Corporation Audit
    path("corporation/add/", add_corp, name="ledger_add_corp"),
    # -- PvE
    path("index", ledger_index, name="ledger_index"),
    # -- -- Corporation Ledger
    path(
        "corporation_ledger/<int:corporation_pk>/",
        corporation_ledger,
        name="corporation_ledger",
    ),
    path("corporation_admin/", corporation_admin, name="corporation_admin"),
    # -- -- Char Ledger
    path(
        "character_ledger/<int:character_pk>/",
        character_ledger,
        name="character_ledger",
    ),
    path("character_admin/", character_admin, name="character_admin"),
    # -- -- Events
    path("events/", events_index, name="events_index"),
    path("events/admin/", events_admin, name="event_admin"),
    path("events/create/", create_event, name="create_event"),
    path("events/<int:event_id>/edit/", edit_event, name="edit_event"),
    path("events/<int:event_id>/delete/", delete_event, name="delete_event"),
    path("events/ajax/load_events", load_events, name="load_events"),
    # -- API System
    re_path(r"^api/", api.urls),
]



from .journal import LedgerJournalApiEndpoints
from .ledger import LedgerApiEndpoints
from .template import LedgerTemplateApiEndpoints


def setup(api):
    LedgerApiEndpoints(api)
    LedgerTemplateApiEndpoints(api)
    LedgerJournalApiEndpoints(api)



# Python
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone

from app_utils.testing import create_user_from_evecharacter

from ledger.models.events import Events
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.views.corporation.corp_events import (
    create_event,
    delete_event,
    edit_event,
    events_admin,
    events_index,
    load_events,
)


class EventViewTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["ledger.basic_access", "ledger.event_admin_access"],
        )
        cls.event = Events.objects.create(
            title="Test Event",
            date_start="2022-01-01T00:00",
            date_end="2030-01-01T01:00",
            description="Test Description",
            char_ledger=False,
            location="Test Location",
        )

    def test_events_index(self):
        self.client.force_login(self.user)
        response = self.client.get("/ledger/events/")
        self.assertEqual(response.status_code, 200)

    def test_create_event(self):
        self.client.force_login(self.user)
        response = self.client.get("/ledger/events/create/")
        self.assertEqual(response.status_code, 200)

    def test_create_event_no_valid(self):
        request = self.factory.post("/ledger/events/create/")
        request.user = self.user
        response = create_event(request)
        self.assertEqual(response.status_code, 200)

    def test_create_event_post(self):
        request = self.factory.post(
            "/ledger/events/create/",
            {
                "title": "New Event",
                "date_start": timezone.now(),
                "date_end": timezone.now() + timezone.timedelta(days=1),
                "location": "New Location",
                "description": "New Description",
                "char_ledger": False,
            },
        )
        request.user = self.user
        response = create_event(request)
        self.assertEqual(response.status_code, 302)

    def test_edit_event_get(self):
        self.client.force_login(self.user)
        response = self.client.get(f"/ledger/events/{self.event.id}/edit/")
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, self.event.title)
        self.assertContains(response, self.event.date_start)
        self.assertContains(response, self.event.date_end)
        self.assertContains(response, self.event.description)
        self.assertContains(response, self.event.location)

    def test_edit_event_post(self):
        self.client.force_login(self.user)
        response = self.client.post(f"/ledger/events/{self.event.id}/edit/")
        self.assertEqual(response.status_code, 200)

    def test_edit_event_post_entry(self):
        self.client.force_login(self.user)
        response = self.client.post(
            f"/ledger/events/{self.event.id}/edit/",
            {
                "title": "Updated Event",
                "date_start": self.event.date_start,
                "date_end": self.event.date_end,
                "location": "Updated Location",
                "description": "Updated Description",
                "char_ledger": self.event.char_ledger,
            },
        )
        self.assertEqual(response.status_code, 302)

    def test_delete_event_get(self):
        self.client.force_login(self.user)
        response = self.client.get(f"/ledger/events/{self.event.id}/delete/")
        self.assertEqual(response.status_code, 200)

    def test_delete_event_post(self):
        request = self.factory.post(f"/ledger/events/{self.event.id}/delete/")
        request.user = self.user
        response = delete_event(request, self.event.id)
        self.assertEqual(response.status_code, 302)

    def test_manage_event(self):
        request = self.factory.post("/ledger/events/admin/")
        request.user = self.user
        response = events_admin(request)
        self.assertEqual(response.status_code, 200)

    def test_load_events(self):
        request = self.factory.get("/ledger/events/load/")
        request.user = self.user
        response = load_events(request)
        self.assertEqual(response.status_code, 200)



from django.test import TestCase

from ledger.templatetags.ledger import ledger_init


class TestTemplateTags(TestCase):
    def test_ledger_init(self):
        result = ledger_init()
        self.assertIsNone(result)



from unittest.mock import patch

from django.test import TestCase

from allianceauth.eveonline.providers import ObjectNotFound

from ledger.models.general import EveEntity
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_eveentity

MODULE_PATH = "ledger.managers.general_manager"


class GeneralQuerySetTest(TestCase):
    @classmethod
    def setUpclass(cls) -> None:
        super().setUpClass()
        load_allianceauth()

    def test_get_or_create_esi(self):
        load_eveentity()
        # given/when
        result = EveEntity.objects.get_or_create_esi(eve_id=1001)
        # then
        self.assertEqual(result[0].eve_id, 1001)

    @patch(MODULE_PATH + ".EveEntityManager.update_or_create_esi")
    def test_get_or_create_esi_doesnotexist(self, mock_update_or_create_esi):
        # given
        mock_update_or_create_esi.return_value = (EveEntity(eve_id=1001), True)
        # when
        result = EveEntity.objects.get_or_create_esi(eve_id=1001)
        # then
        self.assertEqual(result[0].eve_id, 1001)

    @patch(MODULE_PATH + ".esi")
    def test_update_or_create_esi(self, mock_esi):
        # given
        mock_esi.client.Universe.post_universe_names.return_value.results.return_value = [
            {
                "id": 1001,
                "name": "Gneuten",
                "category": "character",
            }
        ]
        # when
        result, _ = EveEntity.objects.update_or_create_esi(eve_id=1001)
        expected_result = EveEntity.objects.get(eve_id=1001)
        # then
        self.assertEqual(result, expected_result)

    @patch(MODULE_PATH + ".esi")
    def test_update_or_create_esi_unknown(self, mock_esi):
        # given
        mock_esi.client.Universe.post_universe_names.return_value.results.return_value = (
            []
        )
        # when
        with self.assertRaises(ObjectNotFound):
            EveEntity.objects.update_or_create_esi(eve_id=1001)

    @patch(MODULE_PATH + ".esi")
    @patch("ledger.models.general.EveEntity.objects.bulk_create")
    def test_create_bulk_from_esi(self, mock_bulk_create, mock_esi):
        # given
        mock_esi.client.Universe.post_universe_names.return_value.results.return_value = [
            {
                "id": 1001,
                "name": "Gneuten",
                "category": "character",
            },
            {
                "id": 1002,
                "name": "rotze Rotineque",
                "category": "character",
            },
        ]
        # when
        result = EveEntity.objects.create_bulk_from_esi([1001, 1002])
        # then
        self.assertTrue(result)
        mock_bulk_create.assert_called_once()

    def test_create_bulk_from_esi_no_ids(self):
        # given
        # when
        result = EveEntity.objects.create_bulk_from_esi([])
        # then
        self.assertTrue(result)



from datetime import timedelta
from unittest.mock import MagicMock, call, patch

from django.db import IntegrityError
from django.test import TestCase
from django.utils import timezone
from esi.errors import TokenExpiredError

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.tasks import (
    create_member_audit,
    create_missing_character,
    update_all_characters,
    update_all_corps,
    update_char_mining_ledger,
    update_char_wallet,
    update_character,
    update_corp,
    update_corp_wallet,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.tasks"


class TestTasks(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.corp_audit_admin_manager",
                "ledger.corp_audit_admin_manager",
            ],
        )
        cls.token = cls.user.token_set.first()
        cls.corporation = cls.character_ownership.character.corporation

    @patch(MODULE_PATH + ".EveCharacter.objects.create_character")
    def test_create_character(self, _):
        # given
        chars_list = [1010, 1011]
        # when
        result = create_missing_character(chars_list=chars_list)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch("memberaudit.models.Character.objects.all")
    @patch(MODULE_PATH + ".CharacterAudit.objects.all")
    def test_create_member_audit(
        self,
        mock_char_audit_all,
        mock_char_all,
        mock_get_character_by_id,
        mock_update_or_create,
    ):
        # Setup mock return values
        mock_char_audit_all.return_value.values_list.return_value = []
        mock_char_all.return_value.values_list.return_value = [1999]
        mock_get_character_by_id.return_value = EveCharacter(
            character_id=1999,
            character_name="Test Character",
            corporation_id=2001,
            corporation_name="Hell Rider'Z",
        )

        # Call the function
        create_member_audit()

        # Check that update_or_create was called
        mock_update_or_create.assert_called_once()

    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch("memberaudit.models.Character.objects.all")
    @patch(MODULE_PATH + ".CharacterAudit.objects.all")
    def test_create_member_audit_no_update_or_create(
        self,
        mock_char_audit_all,
        mock_char_all,
        mock_get_character_by_id,
        mock_update_or_create,
    ):
        # Setup mock return values
        mock_char_audit_all.return_value.values_list.return_value = [1, 2, 3, 4, 5]
        mock_char_all.return_value.values_list.return_value = [2, 3, 4, 5]

        # Call the function
        create_member_audit()

        # Check that update_or_create was not called
        mock_update_or_create.assert_not_called()

    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch("memberaudit.models.Character.objects.all")
    @patch(MODULE_PATH + ".CharacterAudit.objects.all")
    @patch(MODULE_PATH + ".logger")
    def test_create_member_audit_integrityerror(
        self,
        mock_logger,
        mock_char_audit_all,
        mock_char_all,
        mock_get_character_by_id,
        mock_update_or_create,
    ):
        # Setup mock return values
        mock_char_audit_all.return_value.values_list.return_value = [1, 2, 3, 4, 5]
        mock_char_all.return_value.values_list.return_value = [6, 7]
        mock_update_or_create.side_effect = IntegrityError
        # Call the function
        create_member_audit()

        mock_logger.debug.assert_called_once_with(
            "Created %s missing Member Audit Characters", 0
        )

    @patch(MODULE_PATH + ".EveCharacter.objects.create_character")
    def test_create_character_integrity(self, mock_character):
        # given
        chars_list = [1001, 1011]
        mock_character.side_effect = [IntegrityError("duplicate key"), None]
        # when
        result = create_missing_character(chars_list=chars_list)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".update_character.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_characters(self, mock_logger, mock_update_character):
        # given
        characters_count = CharacterAudit.objects.count()
        # when
        update_all_characters()
        # then
        self.assertEqual(mock_update_character.call_count, characters_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Char Audit Updates", characters_count
        )

    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character(self, mock_select_related):
        # given
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now()
        mock_character.last_update_wallet = timezone.now()
        mock_character.character = MagicMock()
        mock_character.character.character_id = self.token.character_id
        mock_character.character.character_name = self.token.character_name

        mock_filter = MagicMock()
        mock_filter.first.return_value = mock_character
        mock_select_related.return_value.filter.return_value = mock_filter
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".update_character_mining")
    def test_update_character_mining(self, mock_char_mining):
        # given
        expected_return_value = f"Finished Mining for: {self.token.character_name}"
        mock_char_mining.return_value = expected_return_value
        # when
        result = update_char_mining_ledger(self.token.character_id)
        # then
        self.assertTrue(mock_char_mining.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_character_wallet")
    def test_update_character_wallet(self, mock_char_wallet):
        # given
        expected_return_value = (
            f"Finished wallet transactions for: {self.token.character_name}"
        )
        mock_char_wallet.return_value = expected_return_value
        # when
        result = update_char_wallet(self.token.character_id)
        # then
        self.assertTrue(mock_char_wallet.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_char_mining_ledger.si")
    @patch(MODULE_PATH + ".update_char_wallet.si")
    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_from_token(
        self,
        mock_check_char,
        mock_update_or_create,
        mock_get_character_by_id,
        mock_get_token,
        mock_char_wallet,
        mock_char_mining,
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = True
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now() - timedelta(days=1)
        mock_character.last_update_wallet = timezone.now() - timedelta(days=1)
        mock_update_or_create.return_value = (mock_character, True)
        # when
        update_character(self.token.character_id)
        # then
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()
        mock_get_character_by_id.assert_called_once_with(mock_token.character_id)
        mock_update_or_create.assert_called_once()
        self.assertTrue(mock_char_wallet.called)
        self.assertTrue(mock_char_mining.called)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_expired_token(
        self, mock_check_char, mock_update_or_create, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.side_effect = TokenExpiredError
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_update_or_create.return_value = (mock_character, True)
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_no_access_token(
        self, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = False
        mock_get_token.return_value = mock_token
        # when
        result = update_character(self.token.character_id)
        # then
        mock_token.valid_access_token.assert_called_once()
        self.assertFalse(result)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_token_no_token(
        self, mock_logger, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_get_token.return_value = False
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_logger.info.assert_called_once_with("No Tokens for %s", 1001)

    @patch(MODULE_PATH + ".update_corp.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_cors(self, mock_logger, mock_update_character):
        # given
        corporation_count = CorporationAudit.objects.count()
        # when
        update_all_corps()
        # then
        self.assertEqual(mock_update_character.call_count, corporation_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Corp Audit Updates", corporation_count
        )

    @patch(MODULE_PATH + ".update_corp_wallet.si")
    def test_update_corp(self, mock_corp_wallet):
        # when
        update_corp(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet.called)

    @patch(MODULE_PATH + ".update_corp_wallet_division")
    def test_update_corp_wallet(self, mock_corp_wallet_division):
        # given
        expected_return_value = (
            f"Finished wallet divs for: {self.corporation.corporation_name}"
        )
        mock_corp_wallet_division.return_value = expected_return_value
        # when
        result = update_corp_wallet(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet_division.called)
        self.assertEqual(expected_return_value, result)



from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.corporation.ledger import LedgerApiEndpoints
from ledger.tests.test_api._ledgercorpdata import noData
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiLedgerCorpEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.corp_audit_manager",
                "ledger.corp_audit_admin_manager",
            ],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
        )

        cls.user3, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    def test_get_corporation_ledger_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2002/ledger/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_ledger_api_year_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/ledger/year/2024/month/0/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_ledger_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/ledger/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_ledger_api_year(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/ledger/year/2024/month/0/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_ledger_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/2001/ledger/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_corporation_ledger_api_no_data(self):
        self.client.force_login(self.user3)
        url = "/ledger/api/corporation/0/ledger/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_ledger_api_not_found(self):
        self.client.force_login(self.user3)
        url = "/ledger/api/corporation/2001/ledger/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")

    @patch("ledger.models.CorporationWalletJournalEntry.objects.filter")
    def test_get_corporation_ledger_api_single_with_zero_summary_amount(
        self, mock_filter
    ):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2002/ledger/year/2024/month/3/"

        # Mock the queryset to return a journal entry with zero summary amount
        mock_entry = MagicMock()
        mock_entry.get.return_value = 0
        mock_entry.get.side_effect = lambda key, default: (
            0 if key in ["total_bounty", "total_ess"] else default
        )
        mock_filter.return_value.select_related.return_value.annotate_ledger.return_value = [
            mock_entry
        ]

        response = self.client.get(url)

        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_admin(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/ledger/admin/"
        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "corporation": {
                    "2002": {"corporation_id": 2002, "corporation_name": "Eulenclub"}
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

    @patch("ledger.api.corporation.ledger.CorporationAudit.objects.visible_to")
    def test_get_corporation_admin_no_visible(self, mock_visible_to):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/ledger/admin/"

        mock_visible_to.return_value = None

        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    @patch("ledger.api.corporation.ledger.CorporationAudit.objects.visible_to")
    def test_get_corporation_admin_exception(self, mock_visible_to):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/ledger/admin/"

        corp = EveCorporationInfo.objects.get(corporation_id=2001)

        mock_visible_to.return_value = [corp, "test"]

        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
